from pyobas.signatures.signature_match import SignatureMatch
from pyobas.signatures.types import MatchTypes, SignatureTypes


class SignatureType:
    """Describes a signature of some time and a matching policy

    :param label: Type specifier
    :type label: SignatureTypes
    :param match_type: the matching policy to use when trying
    to match this signature type, e.g. fuzzy, simple...
    :type match_type: MatchTypes
    :param match_score: if the matching type is fuzzy, this is
    the score to use as threshold, defaults to None
    :type match_score: int, optional
    """

    def __init__(
        self,
        label: SignatureTypes,
        match_type: MatchTypes = MatchTypes.MATCH_TYPE_SIMPLE,
        match_score: int = None,
    ):
        self.label = label
        self.match_policy = SignatureMatch(match_type, match_score)

    def make_struct_for_matching(self, data):
        """Provided some `data`, formats a dictionary specifying the matching
        policy to use by the helper to match expected signatures (from expectations)
        with actual, alert signatures (from the security software)

        :param data: arbitrary data, but most often string or a number primitive
        :type: Any

        :return: dictionary of matching specifiers::
            {
              "type": str,
              "data": any,
              "score": (optional) int
            }
        :rtype: dict
        """
        struct = {
            "type": self.match_policy.match_type.value,
            "data": data,
        }

        if self.match_policy.match_score is not None:
            struct["score"] = self.match_policy.match_score

        return struct



from pyobas.base import RESTManager, RESTObject
from pyobas.mixins import CreateMixin, GetMixin, ListMixin, UpdateMixin
from pyobas.utils import RequiredOptional


class Collector(RESTObject):
    pass


class CollectorManager(GetMixin, ListMixin, CreateMixin, UpdateMixin, RESTManager):
    _path = "/collectors"
    _obj_cls = Collector
    _create_attrs = RequiredOptional(
        required=(
            "collector_id",
            "collector_name",
            "collector_type",
            "collector_period",
        )
    )



from typing import Any, Dict

from pyobas import exceptions as exc
from pyobas.base import RESTManager, RESTObject
from pyobas.mixins import CreateMixin, GetMixin, ListMixin, UpdateMixin
from pyobas.utils import RequiredOptional


class SecurityPlatform(RESTObject):
    pass


class SecurityPlatformManager(
    GetMixin, ListMixin, CreateMixin, UpdateMixin, RESTManager
):
    _path = "/security_platforms"
    _obj_cls = SecurityPlatform
    _create_attrs = RequiredOptional(
        required=("asset_name", "security_platform_type"),
        optional=(
            "asset_description",
            "security_platform_logo_light",
            "security_platform_logo_dark",
        ),
    )

    @exc.on_http_error(exc.OpenBASUpdateError)
    def upsert(
        self, security_platform: Dict[str, Any], **kwargs: Any
    ) -> Dict[str, Any]:
        path = f"{self.path}/upsert"
        result = self.openbas.http_post(path, post_data=security_platform, **kwargs)
        return result



from pyobas.base import RESTManager, RESTObject
from pyobas.mixins import ListMixin, UpdateMixin


class Organization(RESTObject):
    pass


class OrganizationManager(ListMixin, UpdateMixin, RESTManager):
    _path = "/organizations"
    _obj_cls = Organization



from enum import Enum


class MatchTypes(str, Enum):
    MATCH_TYPE_FUZZY = "fuzzy"
    MATCH_TYPE_SIMPLE = "simple"


class SignatureTypes(str, Enum):
    SIG_TYPE_PARENT_PROCESS_NAME = "parent_process_name"
    SIG_TYPE_HOSTNAME = "hostname"
    SIG_TYPE_PROCESS_NAME = "process_name"
    SIG_TYPE_COMMAND_LINE = "command_line"
    SIG_TYPE_FILE_NAME = "file_name"
    SIG_TYPE_IPV4 = "ipv4_address"
    SIG_TYPE_IPV6 = "ipv6_address"



import dataclasses
import datetime
import email.message
import json
import logging
import threading
import urllib.parse
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

import requests
from pythonjsonlogger import jsonlogger


class _StdoutStream:
    def __call__(self, chunk: Any) -> None:
        print(chunk)


def get_content_type(content_type: Optional[str]) -> str:
    message = email.message.Message()
    message["content-type"] = content_type

    return message.get_content_type()


def response_content(
    response: requests.Response,
    streamed: bool,
    action: Optional[Callable[[bytes], None]],
    chunk_size: int,
    *,
    iterator: bool,
) -> Optional[Union[bytes, Iterator[Any]]]:
    if iterator:
        return response.iter_content(chunk_size=chunk_size)

    if streamed is False:
        return response.content

    if action is None:
        action = _StdoutStream()

    for chunk in response.iter_content(chunk_size=chunk_size):
        if chunk:
            action(chunk)
    return None


def copy_dict(
    *,
    src: Dict[str, Any],
    dest: Dict[str, Any],
) -> None:
    for k, v in src.items():
        if isinstance(v, dict):
            for dict_k, dict_v in v.items():
                dest[f"{k}[{dict_k}]"] = dict_v
        else:
            dest[k] = v


class EncodedId(str):
    def __new__(cls, value: Union[str, int, "EncodedId"]) -> "EncodedId":
        if isinstance(value, EncodedId):
            return value

        if not isinstance(value, (int, str)):
            raise TypeError(f"Unsupported type received: {type(value)}")
        if isinstance(value, str):
            value = urllib.parse.quote(value, safe="")
        return super().__new__(cls, value)


def remove_none_from_dict(data: Dict[str, Any]) -> Dict[str, Any]:
    return {k: v for k, v in data.items() if v is not None}


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


@dataclasses.dataclass(frozen=True)
class RequiredOptional:
    required: Tuple[str, ...] = ()
    optional: Tuple[str, ...] = ()
    exclusive: Tuple[str, ...] = ()

    def validate_attrs(
        self,
        *,
        data: Dict[str, Any],
        excludes: Optional[List[str]] = None,
    ) -> None:
        if excludes is None:
            excludes = []

        if self.required:
            required = [k for k in self.required if k not in excludes]
            missing = [attr for attr in required if attr not in data]
            if missing:
                raise AttributeError(f"Missing attributes: {', '.join(missing)}")

        if self.exclusive:
            exclusives = [attr for attr in data if attr in self.exclusive]
            if len(exclusives) > 1:
                raise AttributeError(
                    f"Provide only one of these attributes: {', '.join(exclusives)}"
                )
            if not exclusives:
                raise AttributeError(
                    f"Must provide one of these attributes: "
                    f"{', '.join(self.exclusive)}"
                )


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super(CustomJsonFormatter, self).add_fields(log_record, record, message_dict)
        if not log_record.get("timestamp"):
            # This doesn't use record.created, so it is slightly off
            now = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            log_record["timestamp"] = now
        if log_record.get("level"):
            log_record["level"] = log_record["level"].upper()
        else:
            log_record["level"] = record.levelname


def logger(level, json_logging=True):
    # Exceptions
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("pika").setLevel(logging.ERROR)
    # Exceptions
    if json_logging:
        log_handler = logging.StreamHandler()
        log_handler.setLevel(level)
        formatter = CustomJsonFormatter("%(timestamp)s %(level)s %(name)s %(message)s")
        log_handler.setFormatter(formatter)
        logging.basicConfig(handlers=[log_handler], level=level, force=True)
    else:
        logging.basicConfig(level=level)

    class AppLogger:
        def __init__(self, name):
            self.local_logger = logging.getLogger(name)

        @staticmethod
        def prepare_meta(meta=None):
            return None if meta is None else {"attributes": meta}

        @staticmethod
        def setup_logger_level(lib, log_level):
            logging.getLogger(lib).setLevel(log_level)

        def debug(self, message, meta=None):
            self.local_logger.debug(message, extra=AppLogger.prepare_meta(meta))

        def info(self, message, meta=None):
            self.local_logger.info(message, extra=AppLogger.prepare_meta(meta))

        def warning(self, message, meta=None):
            self.local_logger.warning(message, extra=AppLogger.prepare_meta(meta))

        def error(self, message, meta=None):
            # noinspection PyTypeChecker
            self.local_logger.error(
                message, exc_info=1, extra=AppLogger.prepare_meta(meta)
            )

    return AppLogger


class PingAlive(threading.Thread):
    def __init__(self, api, config, logger, ping_type) -> None:
        threading.Thread.__init__(self)
        self.ping_type = ping_type
        self.api = api
        self.config = config
        self.logger = logger
        self.in_error = False
        self.exit_event = threading.Event()

    def ping(self) -> None:
        while not self.exit_event.is_set():
            try:
                if self.ping_type == "injector":
                    self.api.injector.create(self.config, False)
                else:
                    self.api.collector.create(self.config, False)
            except Exception as err:  # pylint: disable=broad-except
                self.logger.error("Error pinging the API: " + str(err))
            self.exit_event.wait(40)

    def run(self) -> None:
        self.logger.info("Starting PingAlive thread")
        self.ping()

    def stop(self) -> None:
        self.logger.info("Preparing PingAlive for clean shutdown")
        self.exit_event.set()



import abc
import sys
from typing import Any, Dict, Optional, Union

import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder  # type: ignore

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol


class BackendResponse(Protocol):
    @abc.abstractmethod
    def __init__(self, response: requests.Response) -> None: ...


class Backend(Protocol):
    @abc.abstractmethod
    def http_request(
        self,
        method: str,
        url: str,
        json: Optional[Union[Dict[str, Any], bytes]],
        data: Optional[Union[Dict[str, Any], MultipartEncoder]],
        params: Optional[Any],
        timeout: Optional[float],
        verify: Optional[Union[bool, str]],
        stream: Optional[bool],
        **kwargs: Any,
    ) -> BackendResponse: ...



from abc import ABC, abstractmethod
from inspect import signature
from types import FunctionType

from pyobas.client import OpenBAS
from pyobas.configuration import Configuration
from pyobas.exceptions import OpenBASError
from pyobas.utils import logger


class BaseDaemon(ABC):
    """A base class for implementing a kind of daemon that periodically polls
    a given callback.

    :param configuration: configuration to provide the daemon
        (allowing for looking up values within the callback for example)
    :type configuration: Configuration
    :param callback: a method or function to periodically call, defaults to None.
    :type callback: callable, optional
    :param logger: a logger object, to log events. if not supplied, a default logger
        will be spawned to provide this functionality.
    :type logger: Any
    :param api_client: an API client that will provide connectivity with other systems.
    :type api_client: Any
    """

    def __init__(
        self,
        configuration: Configuration,
        callback: callable = None,
        logger=None,
        api_client=None,
    ):
        self._configuration = configuration
        self._callback = callback
        self.api = api_client or BaseDaemon.__get_default_api_client(
            url=self._configuration.get("openbas_url"),
            token=self._configuration.get("openbas_token"),
        )

        # logging
        # compatibility layer: in order for older configs to still work, search for legacy names
        actual_log_level = (
            self._configuration.get("log_level")
            or self._configuration.get("collector_log_level")
            or self._configuration.get("injector_log_level")
            or "info"
        )
        actual_log_name = (
            self._configuration.get("name")
            or self._configuration.get("collector_name")
            or self._configuration.get("injector_name")
            or "daemon"
        )
        self.logger = logger or BaseDaemon.__get_default_logger(
            actual_log_level.upper(),
            actual_log_name,
        )

    @abstractmethod
    def _setup(self):
        """A run-once method that inheritors must implement. This serves to instantiate
        all useful objects and functionality for the implementor to run.
        """
        pass

    @abstractmethod
    def _start_loop(self):
        """Starts the daemon's main execution loop. Implementors should implement
        the main execution logic in here.
        """
        pass

    def _try_callback(self):
        """Tries to call the configured callback. Note that if any error is thrown,
        it is immediately swallowed (but still logged) allowing the collector to keep
        running. This is useful for any transient issue (e.g. API endpoint down...).
        """
        try:
            # this is some black magic to allow injecting the collector daemon instance
            # into an arbitrary callback that has a specific argument name
            # this allow for avoiding subclassing the CollectorDaemon class just to provide the callback
            # Example:
            #
            # def standalone_func(collector):
            #   collector.api.call_openbas()
            #
            # CollectorDaemon(config=<pyboas.configuration.Configuration>, standalone_func).start()
            if (
                isinstance(self._callback, FunctionType)
                and "collector" in signature(self._callback).parameters
            ):
                self._callback(collector=self)
            else:
                self._callback()
        except Exception as err:  # pylint: disable=broad-except
            self.logger.error(f"Error calling: {err}")

    def start(self):
        """Start the daemon. This will run the implementor's run-once setup method and
        follow-up with the main execution loop. Note that at this point, if there is no
        configured callback, the method will abort and kill the daemon.
        """
        if self._callback is None:
            raise OpenBASError("This daemon has no configured callback.")
        self._setup()
        self._start_loop()

    def set_callback(self, callback: callable):
        """Configures a callback to call in the main execution loop. If the callback
        was not provided in the daemon's ctor, this should be set before calling start().
        """
        self._callback = callback

    def get_id(self):
        """Returns the daemon instance's ID contained in configuration. Configuration
        must define any of these keys: `id`, `collector_id`, `injector_id`.
        """
        return (
            self._configuration.get("id")
            or self._configuration.get("collector_id")
            or self._configuration.get("injector_id")
        )

    @classmethod
    def __get_default_api_client(cls, url, token):
        return OpenBAS(url=url, token=token)

    @classmethod
    def __get_default_logger(cls, log_level, name):
        return logger(log_level)(name)



import os


class EnvironmentSource:
    """A utility for fecthing a value in the env vars."""

    @classmethod
    def get(cls, env_var: str) -> str | None:
        """Gets the value for the specified env var

        :param env_var: the name of the env var to query
        :type env_var: str

        :return: value of the env var, or None if not found
        :rtype: str | None
        """
        return os.getenv(env_var)


class DictionarySource:
    """A utility for fetching a value from within a JSON-like (nested dict) structure"""

    # this is quite hacky
    # it only strictly handles two levels of keys in a dict
    @classmethod
    def get(cls, config_key_path: list[str], source_dict: dict) -> str | None:
        """Gets the value for the specified env var

        :param config_key_path: the two-level dictionary path to the config key
        :type config_key_path: list[str]
        :param source_dict: JSON-like (nested dict) structure containing config values.
        :type source_dict: dict

        :return: value for the config key at specified path, or None if not found
        :rtype: str | None
        """
        assert (
            isinstance(config_key_path, list)
            and len(config_key_path) == 2
            and all([len(path_part) > 0 for path_part in config_key_path])
        )
        return source_dict.get(config_key_path[0], {config_key_path[1]: None}).get(
            config_key_path[1]
        )
